@@ -15,8 +15,8 @@ except: from pdb import set_trace as stop # python debugger
 # === START print ===
 # custom print() function override to show caller (module and function) information before actual print string,
 # plus options colourised output
-def print(*args, error=False, warn=False, highlight=False, color="", no_prefix=False, **kwargs):
-    from builtins import print as builtin_print
+def print(*args, error=False, warn=False, highlight=False, color="", input=False, no_prefix=False, **kwargs):
+    from builtins import print as builtin_print, input as builtin_input
     from inspect import currentframe, getouterframes
     from colorama import Fore, Style
     curframe = currentframe()
@@ -52,10 +52,15 @@ def print(*args, error=False, warn=False, highlight=False, color="", no_prefix=F
     if error:
         message = "ERROR: "
         color_str = f"{Fore.RED}"
+    post_str = ""
+    if input: # in case we want to stop here and let the user press key to proceed
+        post_str = " Press Enter to proceed..."
     # print
     builtin_print(color_str+message, end="")
-    builtin_print(*args, **kwargs, flush=True)
-    builtin_print(color_str, end=f"{Style.RESET_ALL}")
+    builtin_print(*args, **kwargs, end="")
+    builtin_print(post_str, end="")
+    builtin_print(color_str, end=f"{Style.RESET_ALL}\n", flush=True)
+    if input: builtin_input() # this will ask for user press key...
     # error handling (exit or stop the code)
     if error:
         if type(error)==str:
@@ -101,7 +106,7 @@ def plot(y=None, x=None, yerr=None, xerr=None, type=None, xlabel='x', ylabel='y'
         if xlog: ax.set_xscale('log')
         if ylog: ax.set_yscale('log')
         legend_handles, legend_labels = ax.get_legend_handles_labels()
-        if legend_labels or legend_handles: plt.legend(loc=legend_loc)
+        if legend_labels or legend_handles: plt.legend(loc=legend_loc, *args, **kwargs)
         if save:
             plt.savefig(save, bbox_inches='tight')
             print(save+' written.', color='magenta')
@@ -112,7 +117,7 @@ def plot(y=None, x=None, yerr=None, xerr=None, type=None, xlabel='x', ylabel='y'
             if pause:
                 plt.draw()
                 plt.pause(pause)
-        plt.clf() # clear figure after use
+        plt.clf(); plt.cla(); plt.close(); plt = None # clear figure after use
     return plt
 # === END plot ===
 
@@ -153,10 +158,35 @@ def plot_map(image=None, dims=None, vmin=None, vmax=None, log=False, colorbar=Tr
             if pause:
                 plt.draw()
                 plt.pause(pause)
-        plt.clf() # clear figure after use
+        plt.clf(); plt.cla(); plt.close(); plt = None # clear figure after use
     return plt
 # === END plot_map ===
 
+# === START rgba2data ===
+# Given a RGBA colour image (2D array) plotted with a colourbar 'cmap', and vmin/vmax values on the colourbar.
+# Return the interpolated data values in the RGBA colour image.
+# Adopted from http://stackoverflow.com/questions/3720840/how-to-reverse-color-map-image-to-scalar-values/3722674#3722674
+def rgba2data(rgba_image, cmap_name, cmap_vmin, cmap_vmax):
+    import matplotlib.pyplot as plt
+    import scipy.cluster.vq as scv
+    # get cmap
+    cmap = plt.get_cmap(cmap_name)
+    gradient = cmap(np.linspace(0.0, 1.0, 100))
+    # Reshape rgba_image with all the 4-tuples in a long list.
+    rgba_image_2 = rgba_image.reshape((rgba_image.shape[0]*rgba_image.shape[1], rgba_image.shape[2]))
+    # Use vector quantization to shift the values in rgba_image_2 to the nearest point in the code book (gradient).
+    code, _ = scv.vq(rgba_image_2, gradient)
+    # 'code' is an array of length rgba_image_2, holding the code book index for each observation (rgba_image_2 are the "observations").
+    # Scale values so they are from 0 to 1.
+    values = code.astype('float') / gradient.shape[0]
+    # Reshape values back to original size of rgba_image
+    values = values.reshape(rgba_image.shape[0], rgba_image.shape[1])
+    values = values[::-1]
+    # Finally, scale output values to cmap range.
+    values = cmap_vmin + values * (cmap_vmax - cmap_vmin)
+    return values
+# === END rgba2data ===
+
 # === START fit ===
 # params should be a dict with the parameter name(s) and a list specifying [min, start, max]
 def fit(func, xdata, ydata, weights=None, params=None, verbose=1, *args, **kwargs):
@@ -164,12 +194,35 @@ def fit(func, xdata, ydata, weights=None, params=None, verbose=1, *args, **kwarg
     model = lmfit_model(func) # get lmfit model object
     # set up parameters
     lmfit_params = model.make_params() # make lmfit default params
-    if params is not None:
+    if params is not None: # user previded parameter bounds and initial values
         for pname in params:
             if pname not in model.param_names: print("parameter key error: '"+pname+"'", error=True)
             plist = params[pname]
             #value = (plist[1]-plist[0]) / 2 # initial parameter value
             model.set_param_hint(pname, min=plist[0], value=plist[1], max=plist[2]) # set bounds for this parameter from [min, val, max]
+    else: # try and find some reasonable initial guesses for the fit parameters
+        if verbose: print("trying to find initial parameter guesses...")
+        # genetic algorithm to come up with initial fit parameter values
+        def generate_initial_params(n_params):
+            from scipy.optimize import differential_evolution
+            # function for genetic algorithm to minimize (sum of squared error)
+            def sum_of_squared_error(parameterTuple):
+                from warnings import filterwarnings
+                filterwarnings("ignore") # do not print warnings by genetic algorithm
+                val = func(xdata, *parameterTuple)
+                return np.sum((ydata - val)**2)
+            # min and max used for bounds
+            maxX = np.max(xdata)
+            maxY = np.max(ydata)
+            maxXY = np.max([maxX, maxY])
+            parameterBounds = [[-maxXY, maxXY]]*n_params
+            result = differential_evolution(sum_of_squared_error, parameterBounds, seed=140281)
+            return result.x
+        # get initial parameter value guesses and put them in
+        initial_params = generate_initial_params(len(model.param_names))
+        for ip, pname in enumerate(model.param_names):
+            model.set_param_hint(pname, value=initial_params[ip])
+    # re-create fit parameter settings after parameter bounds or initial guesses have been determined
     lmfit_params = model.make_params() # update lmfit params
     # get initial parameter info
     if verbose > 1:
@@ -250,7 +303,7 @@ class timer:
     def report(self):
         if self.dt is None: self.get_dt()
         print("timer('"+self.name+"'): start = "+str(self.start_time)+", stop = "+str(self.stop_time)+
-              ", dt = "+str(self.dt), ", dt (seconds) = "+str(self.dt.total_seconds())+" s")
+              ", dt = "+str(self.dt)+", dt (seconds) = "+str(self.dt.total_seconds())+" s", highlight=1)
 
 # === executes a shell command: input string 'cmd'
 def run_shell_command(cmd, quiet=False, print_only=False, capture=False, **kargs):
